import os
from functools import partial
from pathlib import Path
from typing import Any
from typing import Callable
from typing import Dict
from typing import Optional

import pytest
import yaml

from .common import check_text_files
from .common import perform_regression_check


class DataRegressionFixture:
    """
    Implementation of `data_regression` fixture.
    """

    def __init__(
        self, datadir: Path, original_datadir: Path, request: pytest.FixtureRequest
    ) -> None:
        self.request = request
        self.datadir = datadir
        self.original_datadir = original_datadir
        self.force_regen = False
        self.with_test_class_names = False

    def check(
        self,
        data_dict: Dict[str, Any],
        basename: Optional[str] = None,
        fullpath: Optional["os.PathLike[str]"] = None,
    ) -> None:
        """
        Checks the given dict against a previously recorded version, or generate a new file.

        :param data_dict: any yaml serializable dict.

        :param basename: basename of the file to test/record. If not given the name
            of the test is used.
            Use either `basename` or `fullpath`.

        :param fullpath: complete path to use as a reference file. This option
            will ignore ``datadir`` fixture when reading *expected* files but will still use it to
            write *obtained* files. Useful if a reference file is located in the session data dir for example.

        ``basename`` and ``fullpath`` are exclusive.
        """
        __tracebackhide__ = True

        def dump(filename: Path) -> None:
            """Dump dict contents to the given filename"""
            import yaml

            dumped_str = yaml.dump_all(
                [data_dict],
                Dumper=RegressionYamlDumper,
                default_flow_style=False,
                allow_unicode=True,
                indent=2,
                encoding="utf-8",
            )
            with filename.open("wb") as f:
                f.write(dumped_str)

        perform_regression_check(
            datadir=self.datadir,
            original_datadir=self.original_datadir,
            request=self.request,
            check_fn=partial(check_text_files, encoding="UTF-8"),
            dump_fn=dump,
            extension=".yml",
            basename=basename,
            fullpath=fullpath,
            force_regen=self.force_regen,
            with_test_class_names=self.with_test_class_names,
        )

    # non-PEP 8 alias used internally at ESSS
    Check = check


class RegressionYamlDumper(yaml.SafeDumper):
    """
    Custom YAML dumper aimed for regression testing. Differences to usual YAML dumper:

    * Doesn't support aliases, as they produce confusing results on regression tests. The most
    definitive way to get rid of YAML aliases in the dump is to create an specialization that
    never allows aliases, as there isn't an argument that offers same guarantee
    (see http://pyyaml.org/ticket/91).
    """

    def ignore_aliases(self, data: object) -> bool:
        return True

    @classmethod
    def add_custom_yaml_representer(
        cls, data_type: type, representer_fn: Callable[[object, Any], None]
    ) -> None:
        """
        Add custom representer to regression YAML dumper. It is polymorphic, so it works also for
        subclasses of `data_type`.

        :param data_type: Type of objects.
        :param representer_fn: Function that receives ``(dumper, data)`` type as
            argument and must must return a YAML-convertible representation.
        """
        # Use multi-representer instead of simple-representer because it supports polymorphism.
        yaml.add_multi_representer(
            data_type, multi_representer=representer_fn, Dumper=cls
        )



import os
from typing import Any
from typing import Dict
from typing import Optional
from typing import Sequence

from .common import import_error_message
from .dataframe_regression import DataFrameRegressionFixture


class NumericRegressionFixture(DataFrameRegressionFixture):
    """
    Numeric Data Regression fixture implementation used on num_regression fixture.
    """

    def check(
        self,
        data_dict: Dict[str, Any],
        basename: Optional[str] = None,
        fullpath: Optional["os.PathLike[str]"] = None,
        tolerances: Optional[Dict[str, Dict[str, float]]] = None,
        default_tolerance: Optional[Dict[str, float]] = None,
        data_index: Optional[Sequence[int]] = None,
        fill_different_shape_with_nan: bool = True,
    ) -> None:
        """
        Checks the given dict against a previously recorded version, or generate a new file.
        The dict must map from user-defined keys to 1d numpy arrays or array-like values.

        Example::

            num_regression.check({
                'U_gas': U[0][positions],
                'U_liquid': U[1][positions],
                'gas_vol_frac [-]': vol_frac[0][positions],
                'liquid_vol_frac [-]': vol_frac[1][positions],
                'P': Pa_to_bar(P)[positions],
            })

        :param data_dict: dict mapping keys to numpy arrays, or objects that can be
            coerced to 1d numpy arrays with a numeric dtype (e.g. list, tuple, etc).

        :param basename: basename of the file to test/record. If not given the name
            of the test is used.

        :param fullpath: complete path to use as a reference file. This option
            will ignore embed_data completely, being useful if a reference file is located
            in the session data dir for example.

        :param tolerances: dict mapping keys from the data_dict to tolerance settings for the
            given data. Example::

                tolerances={'U': Tolerance(atol=1e-2)}

        :param default_tolerance: dict mapping the default tolerance for the current check
            call. Example::

                default_tolerance=dict(atol=1e-7, rtol=1e-18).

            If not provided, will use defaults from numpy's ``isclose`` function.

        :param data_index: If set, will override the indexes shown in the outputs. Default
            is panda's default, which is ``range(0, len(data))``.

        :param fill_different_shape_with_nan: If set, all the data provided in the data_dict
            that has size lower than the bigger size will be filled with ``np.NaN``, in order to save
            the data in a CSV file.

        ``basename`` and ``fullpath`` are exclusive.
        """

        try:
            import numpy as np
        except ModuleNotFoundError:
            raise ModuleNotFoundError(import_error_message("Numpy"))
        try:
            import pandas as pd
        except ModuleNotFoundError:
            raise ModuleNotFoundError(import_error_message("Pandas"))

        __tracebackhide__ = True

        for k, obj in data_dict.items():
            if not isinstance(obj, np.ndarray):
                arr = np.atleast_1d(np.asarray(obj))
                if np.issubdtype(arr.dtype, np.number):
                    data_dict[k] = arr

        data_shapes1 = []
        for obj in data_dict.values():
            assert type(obj) in [
                np.ndarray
            ], "Only objects that can be coerced to numpy arrays are valid for numeric_data_regression fixture.\n"
            shape = obj.shape

            assert len(shape) == 1, (
                "Only 1D arrays are supported on num_data_regression fixture.\n"
                "Array with shape %s was given.\n" % (shape,)
            )
            data_shapes1.append(shape[0])

        data_shapes2 = np.array(data_shapes1)
        if not np.all(data_shapes2 == data_shapes2[0]):
            if not fill_different_shape_with_nan:
                assert (
                    False
                ), "Data dict with different array lengths will not be accepted. Try setting fill_different_shape_with_nan=True."
            elif len(data_dict) > 1 and not all(
                np.issubdtype(a.dtype, np.floating) for a in data_dict.values()
            ):
                raise TypeError(
                    "Checking multiple arrays with different shapes are not supported for non-float arrays"
                )
            else:
                max_size = max(data_shapes2)
                for k, obj in data_dict.items():
                    new_data = np.empty(shape=(max_size,), dtype=obj.dtype)
                    new_data[: len(obj)] = obj
                    new_data[len(obj) :] = np.nan
                    data_dict[k] = new_data

        data_frame = pd.DataFrame(data_dict, index=data_index)

        DataFrameRegressionFixture.check(
            self, data_frame, basename, fullpath, tolerances, default_tolerance
        )



import pytest


def test_foo(datadir, num_regression):
    """
    Dumb test, just to generate a expected csv file that must be different from the one gathered
    from `TestClass::test_foo`.
    """
    data = {"ones": [1, 1, 1, 1], "zeros": [0, 0, 0, 0]}
    num_regression.check(data)

    expected_filename = f"{test_foo.__name__}.csv"
    obtained_filename = f"{test_foo.__name__}.obtained.csv"
    assert (datadir / expected_filename).exists()
    assert (datadir / obtained_filename).exists()


class TestClass:
    def test_foo(self, datadir, num_regression):
        """
        Since 2.2.1, pytest-regressions use the test class name to compose the name of the datafiles, by default.
        This tests asserts this behavior.
        """
        data = {"twos": [2, 2, 2, 2], "threes": [3, 3, 3, 3]}

        # equivalent to use "--with-test-class-names" parameter
        num_regression._with_test_class_names = True
        num_regression.check(data)

        expected_filename = f"{TestClass.__name__}_{TestClass.test_foo.__name__}.csv"
        obtained_filename = (
            f"{TestClass.__name__}_{TestClass.test_foo.__name__}.obtained.csv"
        )
        assert (datadir / expected_filename).exists()
        assert (datadir / obtained_filename).exists()


class TestClassWithIgnoredName:
    def test_foo(self, datadir, num_regression):
        """
        Specifies to not use the class name to compose the expected data filename. The filename coincides with the
        expected data filename used by `test_foo` function. The regression test then fails because the expected
        data is different.
        """
        data = {"fours": [4, 4, 4, 4], "fives": [5, 5, 5, 5]}

        # Trying to compare this data with `test_foo` expected data.
        with pytest.raises(AssertionError):
            num_regression.check(data)

        # Assert that we have the expected names, without class name
        expected_filename = f"{TestClassWithIgnoredName.test_foo.__name__}.csv"
        obtained_filename = f"{TestClassWithIgnoredName.test_foo.__name__}.obtained.csv"
        assert (datadir / expected_filename).exists()
        assert (datadir / obtained_filename).exists()

        # Assert that don't have files with class name
        expected_with_class_name = f"{TestClassWithIgnoredName.__name__}_{TestClassWithIgnoredName.test_foo.__name__}.csv"
        obtained_with_class_name = f"{TestClassWithIgnoredName.__name__}_{TestClassWithIgnoredName.test_foo.__name__}.obtained.csv"
        assert not (datadir / expected_with_class_name).exists()
        assert not (datadir / obtained_with_class_name).exists()



import os
import zipfile
from pathlib import Path
from typing import Any
from typing import Dict
from typing import Optional

import pytest

from .common import import_error_message
from .common import perform_regression_check


class NDArraysRegressionFixture:
    """
    NumPy NPZ regression fixture implementation used on ndarrays_regression fixture.
    """

    THRESHOLD = 100
    ROWFORMAT = "{:>15s}  {:>20s}  {:>20s}  {:>20s}\n"

    def __init__(
        self, datadir: Path, original_datadir: Path, request: pytest.FixtureRequest
    ) -> None:
        self._tolerances_dict: Dict[str, Dict[str, float]] = {}
        self._default_tolerance: Dict[str, float] = {}

        self.request = request
        self.datadir = datadir
        self.original_datadir = original_datadir
        self._force_regen = False
        self._with_test_class_names = False

    def _check_data_types(
        self, key: Any, obtained_array: Any, expected_array: Any
    ) -> None:
        """
        Check if data type of obtained and expected arrays are the same. Fail if not.
        Helper method used in _check_fn method.
        """
        try:
            import numpy as np
        except ModuleNotFoundError:
            raise ModuleNotFoundError(import_error_message("NumPy"))

        __tracebackhide__ = True

        if obtained_array.dtype != expected_array.dtype:
            # Check if both data types are comparable as numbers (float, int, short, bytes, etc...)
            if np.issubdtype(obtained_array.dtype, np.number) and np.issubdtype(
                expected_array.dtype, np.number
            ):
                return
            # Check if both data types are comparable as strings
            if np.issubdtype(obtained_array.dtype, str) and np.issubdtype(
                expected_array.dtype, str
            ):
                return

            # In case they are not, assume they are not comparable
            error_msg = (
                "Data types are not the same.\n"
                f"  key: {key}\n"
                f"  Obtained: {obtained_array.dtype}\n"
                f"  Expected: {expected_array.dtype}\n"
            )
            raise AssertionError(error_msg)

    def _check_data_shapes(
        self, key: Any, obtained_array: Any, expected_array: Any
    ) -> None:
        """
        Check if obtained and expected arrays have the same size.
        Helper method used in _check_fn method.
        """
        __tracebackhide__ = True

        if obtained_array.shape != expected_array.shape:
            error_msg = (
                "Shapes are not the same.\n"
                f"  Key: {key}\n"
                f"  Obtained: {obtained_array.shape}\n"
                f"  Expected: {expected_array.shape}\n"
            )
            raise AssertionError(error_msg)

    def _check_fn(self, obtained_filename: Path, expected_filename: Path) -> None:
        """
        Check if dict contents dumped to a file match the contents in expected file.
        """
        try:
            import numpy as np
        except ModuleNotFoundError:
            raise ModuleNotFoundError(import_error_message("NumPy"))

        __tracebackhide__ = True

        # Turn result of np.load into a dictionary, such that the files are closed immediately.
        expected_data = self._load_fn(expected_filename)
        obtained_data = self._load_fn(obtained_filename)

        # Check mismatches in the keys.
        if set(obtained_data) != set(expected_data):
            error_msg = (
                "They keys in the obtained results differ from the expected results.\n"
            )
            error_msg += "  Matching keys: "
            error_msg += str(list(set(obtained_data) & set(expected_data)))
            error_msg += "\n"
            error_msg += "  New in obtained: "
            error_msg += str(list(set(obtained_data) - set(expected_data)))
            error_msg += "\n"
            error_msg += "  Missing from obtained: "
            error_msg += str(list(set(expected_data) - set(obtained_data)))
            error_msg += "\n"
            error_msg += "To update values, use --force-regen option.\n\n"
            raise AssertionError(error_msg)

        # Compare the contents of the arrays.
        comparison_tables_dict = {}
        for k, obtained_array in obtained_data.items():
            expected_array = expected_data[k]
            tolerance_args = self._tolerances_dict.get(k, self._default_tolerance)

            self._check_data_types(k, obtained_array, expected_array)
            self._check_data_shapes(k, obtained_array, expected_array)

            if np.issubdtype(obtained_array.dtype, np.inexact):
                not_close_mask = ~np.isclose(
                    obtained_array,
                    expected_array,
                    equal_nan=True,
                    **tolerance_args,
                )
            else:
                not_close_mask = obtained_array != expected_array

            if np.any(not_close_mask):
                diff_ids: Any
                if not_close_mask.ndim == 0:
                    diff_ids = [()]
                else:
                    diff_ids = np.array(
                        np.nonzero(not_close_mask)
                    ).T  # type:ignore[unreachable]
                comparison_tables_dict[k] = (
                    expected_array.size,
                    expected_array.shape,
                    diff_ids,
                    obtained_array[not_close_mask],
                    expected_array[not_close_mask],
                )

        if len(comparison_tables_dict) > 0:
            error_msg = "Values are not sufficiently close.\n"
            error_msg += "To update values, use --force-regen option.\n\n"
            for k, (
                size,
                shape,
                diff_ids,
                obtained_array,
                expected_array,
            ) in comparison_tables_dict.items():
                # Summary
                error_msg += f"{k}:\n  Shape: {shape}\n"
                pct = 100 * len(diff_ids) / size
                error_msg += (
                    f"  Number of differences: {len(diff_ids)} / {size} ({pct:.1f}%)\n"
                )
                if np.issubdtype(obtained_array.dtype, np.number) and len(diff_ids) > 1:
                    error_msg += (
                        "  Statistics are computed for differing elements only.\n"
                    )

                    abs_errors = abs(obtained_array - expected_array)
                    error_msg += "  Stats for abs(obtained - expected):\n"
                    error_msg += f"    Max:     {abs_errors.max()}\n"
                    error_msg += f"    Mean:    {abs_errors.mean()}\n"
                    error_msg += f"    Median:  {np.median(abs_errors)}\n"

                    expected_nonzero = np.array(np.nonzero(expected_array)).T
                    rel_errors = abs(
                        (
                            obtained_array[expected_nonzero]
                            - expected_array[expected_nonzero]
                        )
                        / expected_array[expected_nonzero]
                    )
                    if len(rel_errors) == 0:
                        error_msg += "  Relative errors are not reported because all expected values are zero.\n"
                    else:
                        error_msg += (
                            f"  Stats for abs(obtained - expected) / abs(expected):\n"
                        )
                        if len(rel_errors) != len(abs_errors):
                            pct = 100 * len(rel_errors) / len(abs_errors)
                            error_msg += f"    Number of (differing) non-zero expected results: {len(rel_errors)} / {len(abs_errors)} ({pct:.1f}%)\n"
                            error_msg += f"    Relative errors are computed for the non-zero expected results.\n"
                        else:
                            rel_errors = abs(
                                (obtained_array - expected_array) / expected_array
                            )
                        error_msg += f"    Max:     {rel_errors.max()}\n"
                        error_msg += f"    Mean:    {rel_errors.mean()}\n"
                        error_msg += f"    Median:  {np.median(rel_errors)}\n"

                # Details results
                error_msg += "  Individual errors:\n"
                if len(diff_ids) > self.THRESHOLD:
                    error_msg += (
                        f"    Only showing first {self.THRESHOLD} mismatches.\n"
                    )
                    diff_ids = diff_ids[: self.THRESHOLD]
                    obtained_array = obtained_array[: self.THRESHOLD]
                    expected_array = expected_array[: self.THRESHOLD]
                error_msg += self.ROWFORMAT.format(
                    "Index",
                    "Obtained",
                    "Expected",
                    "Difference",
                )
                for diff_id, obtained, expected in zip(
                    diff_ids, obtained_array, expected_array
                ):
                    diff_id_str = ", ".join(str(i) for i in diff_id)
                    if len(diff_id) != 1:
                        diff_id_str = f"({diff_id_str})"
                    error_msg += self.ROWFORMAT.format(
                        diff_id_str,
                        str(obtained),
                        str(expected),
                        str(obtained - expected)
                        if isinstance(obtained, np.number)
                        else "",
                    )
                error_msg += "\n"

            raise AssertionError(error_msg)

    def _load_fn(self, filename: Path) -> Dict[str, Any]:
        """
        Load dict contents from the given filename.
        """
        try:
            import numpy as np
        except ModuleNotFoundError:
            raise ModuleNotFoundError(import_error_message("NumPy"))

        try:
            # Open the file with a context manager manually, because np.load does not
            # follow such good practices internally, causing avoidable error messages
            # in the unit tests.
            with open(filename, "rb") as f:
                result = dict(np.load(f))
        except (zipfile.BadZipFile, ValueError) as e:
            raise IOError(
                f"NPZ file {filename} could not be loaded. Corrupt file?"
            ) from e
        return result

    def _dump_fn(self, data_dict: Dict[str, Any], filename: Path) -> None:
        """
        Dump dict contents to the given filename.
        """
        try:
            import numpy as np
        except ModuleNotFoundError:
            raise ModuleNotFoundError(import_error_message("NumPy"))

        np.savez_compressed(str(filename), **data_dict)

    def check(
        self,
        data_dict: Dict[str, Any],
        basename: Optional[str] = None,
        fullpath: Optional["os.PathLike[str]"] = None,
        tolerances: Optional[Dict[str, Dict[str, float]]] = None,
        default_tolerance: Optional[Dict[str, float]] = None,
    ) -> None:
        """
        Checks a dictionary of NumPy ndarrays, containing only numeric data, against a previously recorded version, or generate a new file.

        Example::

            def test_some_data(ndarrays_regression):
                points, values = some_function()
                ndarrays_regression.check(
                    {
                        'points': points,  # array with shape (100, 3)
                        'values': values,  # array with shape (100,)
                    },
                    default_tolerance=dict(atol=1e-8, rtol=1e-8)
                )

        :param data_dict: dictionary of NumPy ndarrays containing
            data for regression check. The arrays can have any shape.

        :param basename: basename of the file to test/record. If not given the name
            of the test is used.

        :param fullpath: complete path to use as a reference file. This option
            will ignore embed_data completely, being useful if a reference file is located
            in the session data dir for example.

        :param tolerances: dict mapping keys from the data_dict to tolerance settings
            for the given data. Example::

                tolerances={'U': Tolerance(atol=1e-2)}

        :param default_tolerance: dict mapping the default tolerance for the current
            check call. Example::

                default_tolerance=dict(atol=1e-7, rtol=1e-18).

            If not provided, will use defaults from numpy's ``isclose`` function.

        ``basename`` and ``fullpath`` are exclusive.
        """
        try:
            import numpy as np
        except ModuleNotFoundError:
            raise ModuleNotFoundError(import_error_message("NumPy"))

        import functools

        __tracebackhide__ = True

        if not isinstance(data_dict, dict):
            raise TypeError(
                "Only dictionaries with NumPy arrays or array-like objects are "
                "supported on ndarray_regression fixture.\n"
                "Object with type '{}' was given.".format(str(type(data_dict)))
            )
        for key, array in data_dict.items():
            assert isinstance(key, str), (
                "The dictionary keys must be strings. "
                "Found key with type '{}'".format(str(type(key)))
            )
            data_dict[key] = np.asarray(array)

        for key, array in data_dict.items():
            # Accepted:
            #  - b: boolean
            #  - i: signed integer
            #  - u: unsigned integer
            #  - f: floating-point number
            #  - c: complex floating-point number
            #  - U: unicode string
            # Rejected:
            #  - m: timedelta
            #  - M: datetime
            #  - O: objects
            #  - S: zero-terminated bytes
            #  - V: void (raw data, structured arrays)
            if array.dtype.kind not in ["b", "i", "u", "f", "c", "U"]:
                raise TypeError(
                    "Only numeric or unicode data is supported on ndarrays_regression "
                    f"fixture.\nArray '{key}' with type '{array.dtype}' was given."
                )

        if tolerances is None:
            tolerances = {}
        self._tolerances_dict = tolerances

        if default_tolerance is None:
            default_tolerance = {}
        self._default_tolerance = default_tolerance

        dump_fn = functools.partial(self._dump_fn, data_dict)

        perform_regression_check(
            datadir=self.datadir,
            original_datadir=self.original_datadir,
            request=self.request,
            check_fn=self._check_fn,
            dump_fn=dump_fn,
            extension=".npz",
            basename=basename,
            fullpath=fullpath,
            force_regen=self._force_regen,
            with_test_class_names=self._with_test_class_names,
        )



from pathlib import Path
from typing import Any
from typing import TYPE_CHECKING

import pytest

if TYPE_CHECKING:
    from .data_regression import DataRegressionFixture
    from .dataframe_regression import DataFrameRegressionFixture
    from .ndarrays_regression import NDArraysRegressionFixture
    from .file_regression import FileRegressionFixture
    from .num_regression import NumericRegressionFixture
    from .image_regression import ImageRegressionFixture


def pytest_addoption(parser: Any) -> None:
    group = parser.getgroup("regressions")
    group.addoption(
        "--force-regen",
        action="store_true",
        default=False,
        help="Regenerate regression data files, failing tests with different data.",
    )
    group.addoption(
        "--regen-all",
        action="store_true",
        default=False,
        help="Regenerate all files, letting tests pass (use to regenerate everything in one run).",
    )
    group.addoption(
        "--with-test-class-names",
        action="store_true",
        default=False,
        help="Do not ignore the names of the test classes when composing the name of the regression data files.",
    )


@pytest.fixture
def data_regression(
    datadir: Path, original_datadir: Path, request: pytest.FixtureRequest
) -> "DataRegressionFixture":
    """
    Fixture used to test arbitrary data against known versions previously
    recorded by this same fixture. Useful to test 3rd party APIs or where testing directly
    generated data from other sources.

    Create a dict in your test containing a arbitrary data you want to test, and
    call the `Check` function. The first time it will fail but will generate a file in your
    data directory.

    Subsequent runs against the same data will now compare against the generated file and pass
    if the dicts are equal, or fail showing a diff otherwise. To regenerate data for failed tests,
    either set `force_regen` attribute to True or pass the `--force-regen` flag to pytest
    which will regenerate the data files for all failed tests. Make sure to diff the files to ensure
    the new changes are expected and not regressions.

    The dict may be anything serializable by the `yaml` library.
    """
    from .data_regression import DataRegressionFixture

    return DataRegressionFixture(datadir, original_datadir, request)


@pytest.fixture
def dataframe_regression(
    datadir: Path, original_datadir: Path, request: pytest.FixtureRequest
) -> "DataFrameRegressionFixture":
    """
    Example usage:

    def testSomeData(dataframe_regression):
        dataframe_regression.check(
            pandas.DataFrame.from_dict(
                {
                    'U_gas': U[0],
                    'U_liquid': U[1],
                    'gas_vol_frac [-]': α[0],
                    'liquid_vol_frac [-]': α[1],
                    'P': Pa_to_bar(P),
                }
            ),
            default_tolerance=dict(atol=1e-8, rtol=1e-8)
        )
    """
    from .dataframe_regression import DataFrameRegressionFixture

    return DataFrameRegressionFixture(datadir, original_datadir, request)


@pytest.fixture
def ndarrays_regression(
    datadir: Path, original_datadir: Path, request: pytest.FixtureRequest
) -> "NDArraysRegressionFixture":
    """
    Similar to num_regression, but supports numpy arrays with arbitrary shape. The
    dictionary is stored as an NPZ file. The values of the dictionary must be accepted
    by ``np.asarray``.

    Example::

        def test_some_data(ndarrays_regression):
            points, values = some_function()
            ndarrays_regression.check(
                {
                    'points': points,  # array with shape (100, 3)
                    'values': values,  # array with shape (100,)
                },
                default_tolerance=dict(atol=1e-8, rtol=1e-8)
            )
    """
    from .ndarrays_regression import NDArraysRegressionFixture

    return NDArraysRegressionFixture(datadir, original_datadir, request)


@pytest.fixture
def file_regression(
    datadir: Path, original_datadir: Path, request: pytest.FixtureRequest
) -> "FileRegressionFixture":
    """
    Very similar to `data_regression`, but instead of saving data to YAML file it saves to an
    arbitrary format.

    Useful when used to compare contents of files of specific format (like documents, for instance).
    """
    from .file_regression import FileRegressionFixture

    return FileRegressionFixture(datadir, original_datadir, request)


@pytest.fixture
def num_regression(
    datadir: Path, original_datadir: Path, request: pytest.FixtureRequest
) -> "NumericRegressionFixture":
    """
    Example usage:

    def testSomeData(num_regression):
        num_regression.check(
            {
                'U_gas': U[0],
                'U_liquid': U[1],
                'gas_vol_frac [-]': α[0],
                'liquid_vol_frac [-]': α[1],
                'P': Pa_to_bar(P),
            },
            data_index=positions,
            default_tolerance=dict(atol=1e-8, rtol=1e-8)
        )
    """
    from .num_regression import NumericRegressionFixture

    return NumericRegressionFixture(datadir, original_datadir, request)


@pytest.fixture
def image_regression(
    datadir: Path, original_datadir: Path, request: pytest.FixtureRequest
) -> "ImageRegressionFixture":
    """
    Regression checks for images, accounting for small differences between them.

    Example:
        def test_plots(datadir, image_regression):
            path = generate_plot(datadir / 'plot.png')
            image_regression.check(path.read_bytes(), diff_threshold=2.5)  # 2.5%
    """
    from .image_regression import ImageRegressionFixture

    return ImageRegressionFixture(datadir, original_datadir, request)



import sys

import numpy as np
import pytest

from pytest_regressions.testing import check_regression_fixture_workflow


@pytest.fixture
def no_regen(ndarrays_regression, request):
    if ndarrays_regression._force_regen or request.config.getoption("force_regen"):
        pytest.fail("--force-regen should not be used on this test.")


def test_usage_workflow(pytester, monkeypatch):
    monkeypatch.setattr(
        sys, "testing_get_data", lambda: {"data": 1.1 * np.ones(50)}, raising=False
    )
    source = """
        import sys
        def test_1(ndarrays_regression):
            contents = sys.testing_get_data()
            ndarrays_regression.check(contents)
    """

    def get_npz_contents():
        filename = pytester.path / "test_file" / "test_1.npz"
        return dict(np.load(str(filename)))

    def compare_arrays(obtained, expected):
        assert (obtained["data"] == expected["data"]).all()

    check_regression_fixture_workflow(
        pytester,
        source=source,
        data_getter=get_npz_contents,
        data_modifier=lambda: monkeypatch.setattr(
            sys, "testing_get_data", lambda: {"data": 1.2 * np.ones(50)}, raising=False
        ),
        expected_data_1={"data": 1.1 * np.ones(50)},
        expected_data_2={"data": 1.2 * np.ones(50)},
        compare_fn=compare_arrays,
    )


def test_common_case(ndarrays_regression, no_regen):
    # Most common case: Data is valid, is present and should pass
    data1 = np.full(5000, 1.1, dtype=float)
    data2 = np.arange(5000, dtype=int)
    ndarrays_regression.check({"data1": data1, "data2": data2})

    # Assertion error case 1: Data has one invalid place
    data1 = np.full(5000, 1.1, dtype=float)
    data2 = np.arange(5000, dtype=int)
    data1[500] += 0.1
    with pytest.raises(AssertionError) as excinfo:
        ndarrays_regression.check({"data1": data1, "data2": data2})
    obtained_error_msg = str(excinfo.value)
    expected = "\n".join(
        [
            "Values are not sufficiently close.",
            "To update values, use --force-regen option.",
        ]
    )
    assert expected in obtained_error_msg
    expected = "\n".join(
        [
            "data1:",
            "  Shape: (5000,)",
            "  Number of differences: 1 / 5000 (0.0%)",
            "  Individual errors:",
            "          Index              Obtained              Expected            Difference",
            "            500    1.2000000000000002                   1.1   0.10000000000000009",
        ]
    )
    assert expected in obtained_error_msg

    # Assertion error case 2: More than one invalid data
    data1 = np.full(5000, 1.1, dtype=float)
    data2 = np.arange(5000, dtype=int)
    data1[500] += 0.1
    data1[600] += 0.2
    data2[0] += 5
    data2[700:900] += 5
    with pytest.raises(AssertionError) as excinfo:
        ndarrays_regression.check({"data1": data1, "data2": data2})
    obtained_error_msg = str(excinfo.value)
    expected = "\n".join(
        [
            "Values are not sufficiently close.",
            "To update values, use --force-regen option.",
        ]
    )
    assert expected in obtained_error_msg
    expected = "\n".join(
        [
            "data1:",
            "  Shape: (5000,)",
            "  Number of differences: 2 / 5000 (0.0%)",
            "  Statistics are computed for differing elements only.",
            "  Stats for abs(obtained - expected):",
            "    Max:     0.19999999999999996",
            "    Mean:    0.15000000000000002",
            "    Median:  0.15000000000000002",
            "  Stats for abs(obtained - expected) / abs(expected):",
            "    Max:     0.18181818181818177",
            "    Mean:    0.13636363636363638",
            "    Median:  0.13636363636363638",
            "  Individual errors:",
            "          Index              Obtained              Expected            Difference",
            "            500    1.2000000000000002                   1.1   0.10000000000000009",
            "            600                   1.3                   1.1   0.19999999999999996",
        ]
    )
    assert expected in obtained_error_msg
    expected = "\n".join(
        [
            "data2:",
            "  Shape: (5000,)",
            "  Number of differences: 201 / 5000 (4.0%)",
            "  Statistics are computed for differing elements only.",
            "  Stats for abs(obtained - expected):",
            "    Max:     5",
            "    Mean:    5.0",
            "    Median:  5.0",
            "  Stats for abs(obtained - expected) / abs(expected):",
            "    Number of (differing) non-zero expected results: 200 / 201 (99.5%)",
            "    Relative errors are computed for the non-zero expected results.",
            "    Max:     0.007142857142857143",
            "    Mean:    0.006286830640674575",
            "    Median:  0.006253911138923655",
            "  Individual errors:",
            "    Only showing first 100 mismatches.",
            "          Index              Obtained              Expected            Difference",
            "              0                     5                     0                     5",
            "            700                   705                   700                     5",
            "            701                   706                   701                     5",
        ]
    )
    assert expected in obtained_error_msg


def test_common_case_nd(ndarrays_regression, no_regen):
    # Most common case: Data is valid, is present and should pass
    data1 = np.full((50, 20), 1.1, dtype=float)
    data2 = np.arange(60, dtype=int).reshape((3, 4, 5))
    ndarrays_regression.check({"data1": data1, "data2": data2})

    # Assertion error case 1: Data has one invalid place
    data1 = np.full((50, 20), 1.1, dtype=float)
    data2 = np.arange(60, dtype=int).reshape((3, 4, 5))
    data1[30, 2] += 0.1
    with pytest.raises(AssertionError) as excinfo:
        ndarrays_regression.check({"data1": data1, "data2": data2})
    obtained_error_msg = str(excinfo.value)
    expected = "\n".join(
        [
            "Values are not sufficiently close.",
            "To update values, use --force-regen option.",
        ]
    )
    assert expected in obtained_error_msg
    expected = "\n".join(
        [
            "data1:",
            "  Shape: (50, 20)",
            "  Number of differences: 1 / 1000 (0.1%)",
            "  Individual errors:",
            "          Index              Obtained              Expected            Difference",
            "        (30, 2)    1.2000000000000002                   1.1   0.10000000000000009",
        ]
    )
    assert expected in obtained_error_msg

    # Assertion error case 2: More than one invalid data
    data1 = np.full((50, 20), 1.1, dtype=float)
    data2 = np.arange(60, dtype=int).reshape((3, 4, 5))
    data1[20, 15] += 0.1
    data1[0, 9] = 1.43248324e35
    data2[:2, 0, [0, 2, 4]] += 71
    with pytest.raises(AssertionError) as excinfo:
        ndarrays_regression.check({"data1": data1, "data2": data2})
    obtained_error_msg = str(excinfo.value)
    expected = "\n".join(
        [
            "Values are not sufficiently close.",
            "To update values, use --force-regen option.",
        ]
    )
    assert expected in obtained_error_msg
    expected = "\n".join(
        [
            "data1:",
            "  Shape: (50, 20)",
            "  Number of differences: 2 / 1000 (0.2%)",
            "  Statistics are computed for differing elements only.",
            "  Stats for abs(obtained - expected):",
            "    Max:     1.43248324e+35",
            "    Mean:    7.1624162e+34",
            "    Median:  7.1624162e+34",
            "  Stats for abs(obtained - expected) / abs(expected):",
            "    Max:     1.3022574909090907e+35",
            "    Mean:    6.511287454545454e+34",
            "    Median:  6.511287454545454e+34",
            "  Individual errors:",
            "          Index              Obtained              Expected            Difference",
            "         (0, 9)        1.43248324e+35                   1.1        1.43248324e+35",
            "       (20, 15)    1.2000000000000002                   1.1   0.10000000000000009",
        ]
    )
    assert expected in obtained_error_msg
    expected = "\n".join(
        [
            "data2:",
            "  Shape: (3, 4, 5)",
            "  Number of differences: 6 / 60 (10.0%)",
            "  Statistics are computed for differing elements only.",
            "  Stats for abs(obtained - expected):",
            "    Max:     71",
            "    Mean:    71.0",
            "    Median:  71.0",
            "  Stats for abs(obtained - expected) / abs(expected):",
            "    Number of (differing) non-zero expected results: 5 / 6 (83.3%)",
            "    Relative errors are computed for the non-zero expected results.",
            "    Max:     35.5",
            "    Mean:    12.597121212121213",
            "    Median:  3.55",
            "  Individual errors:",
            "          Index              Obtained              Expected            Difference",
            "      (0, 0, 0)                    71                     0                    71",
            "      (0, 0, 2)                    73                     2                    71",
            "      (0, 0, 4)                    75                     4                    71",
            "      (1, 0, 0)                    91                    20                    71",
            "      (1, 0, 2)                    93                    22                    71",
            "      (1, 0, 4)                    95                    24                    71",
        ]
    )
    assert expected in obtained_error_msg


def test_common_case_zero_expected(ndarrays_regression, no_regen):
    # Most common case: Data is valid, is present and should pass
    data = {"data1": np.array([0, 0, 2, 3, 0, 5, 0, 7])}
    ndarrays_regression.check(data)

    # Assertion error case: Only some zeros are not reproduced.
    data = {"data1": np.array([1, 5, 2, 3, 0, 5, 3, 7])}
    with pytest.raises(AssertionError) as excinfo:
        ndarrays_regression.check(data)
    obtained_error_msg = str(excinfo.value)
    expected = "\n".join(
        [
            "Values are not sufficiently close.",
            "To update values, use --force-regen option.",
        ]
    )
    assert expected in obtained_error_msg
    expected = "\n".join(
        [
            "data1:",
            "  Shape: (8,)",
            "  Number of differences: 3 / 8 (37.5%)",
            "  Statistics are computed for differing elements only.",
            "  Stats for abs(obtained - expected):",
            "    Max:     5",
            "    Mean:    3.0",
            "    Median:  3.0",
            "  Relative errors are not reported because all expected values are zero.",
            "  Individual errors:",
            "          Index              Obtained              Expected            Difference",
            "              0                     1                     0                     1",
            "              1                     5                     0                     5",
            "              6                     3                     0                     3",
        ]
    )
    assert expected in obtained_error_msg


def test_different_data_types(ndarrays_regression, no_regen):
    # Generate data with integer array.
    data = {"data1": np.array([1] * 10)}
    ndarrays_regression.check(data)

    # Run check with incompatible type.
    data = {"data1": np.array([True] * 10)}
    with pytest.raises(AssertionError) as excinfo:
        ndarrays_regression.check(data)
    obtained_error_msg = str(excinfo.value)
    expected = "\n".join(
        [
            "Data types are not the same.",
            "  key: data1",
            "  Obtained: bool",
            "  Expected: int64",
        ]
    )
    assert expected in obtained_error_msg


class Foo:
    def __init__(self, bar):
        self.bar = bar


def test_object_dtype(ndarrays_regression, no_regen):
    data1 = {"data1": np.array([Foo(i) for i in range(4)], dtype=object)}
    with pytest.raises(TypeError) as excinfo:
        ndarrays_regression.check(data1)
    obtained_error_msg = str(excinfo.value)
    expected = "\n".join(
        [
            "Only numeric or unicode data is supported on ndarrays_regression fixture.",
            "Array 'data1' with type 'object' was given.",
        ]
    )
    assert expected in obtained_error_msg


def test_integer_values_smoke_test(ndarrays_regression, no_regen):
    data1 = np.ones(11, dtype=int)
    ndarrays_regression.check({"data1": data1})


def test_float_values_smoke_test(ndarrays_regression):
    data1 = np.array([1.2345678e50, 1.2345678e-50, 0.0])
    ndarrays_regression.check({"data1": data1})


def test_bool_array(ndarrays_regression, no_regen):
    # Correct data
    data1 = np.array([False, False, False], dtype=bool)
    ndarrays_regression.check({"data1": data1})

    # Data with errors
    data1 = np.array([True, True, False], dtype=bool)
    with pytest.raises(AssertionError) as excinfo:
        ndarrays_regression.check({"data1": data1})
    obtained_error_msg = str(excinfo.value)
    expected = "\n".join(
        [
            "Values are not sufficiently close.",
            "To update values, use --force-regen option.",
        ]
    )
    assert expected in obtained_error_msg
    expected = "\n".join(
        [
            "data1:",
            "  Shape: (3,)",
            "  Number of differences: 2 / 3 (66.7%)",
            "  Individual errors:",
            "          Index              Obtained              Expected            Difference",
            "              0                  True                 False                      ",
            "              1                  True                 False                      ",
        ]
    )
    assert expected in obtained_error_msg


def test_complex_array(ndarrays_regression, no_regen):
    # Correct data
    data1 = np.array([3.0 + 2.5j, -0.5, -1.879j])
    ndarrays_regression.check({"data1": data1})

    # Data with errors
    data1 = np.array([3.0 + 2.5j, 0.5, -1.879])
    with pytest.raises(AssertionError) as excinfo:
        ndarrays_regression.check({"data1": data1})
    str(excinfo.value)
    expected = "\n".join(
        [
            "data1:",
            "  Shape: (3,)",
            "  Number of differences: 2 / 3 (66.7%)",
            "  Statistics are computed for differing elements only.",
            "  Stats for abs(obtained - expected):",
            "    Max:     2.6573072836990455",
            "    Mean:    1.8286536418495227",
            "    Median:  1.8286536418495227",
            "  Stats for abs(obtained - expected) / abs(expected):",
            "    Max:     2.0",
            "    Mean:    1.7071067811865475",
            "    Median:  1.7071067811865475",
            "  Individual errors:",
            "          Index              Obtained              Expected            Difference",
            "              1              (0.5+0j)             (-0.5+0j)                (1+0j)",
            "              2           (-1.879+0j)           (-0-1.879j)       (-1.879+1.879j)",
        ]
    )


def test_arrays_of_same_size_1d(ndarrays_regression):
    data = {
        "hello": np.zeros((1,), dtype=int),
        "world": np.zeros((1,), dtype=int),
    }
    ndarrays_regression.check(data)


def test_arrays_with_different_sizes_1d(ndarrays_regression, no_regen):
    data = {"data1": np.ones(11, dtype=np.float64)}
    ndarrays_regression.check(data)

    # Original NPY file contains 11 elements.
    data = {"data1": np.ones(10, dtype=np.float64)}
    with pytest.raises(AssertionError) as excinfo:
        ndarrays_regression.check(data)
    obtained_error_msg = str(excinfo.value)
    expected = "\n".join(
        [
            "Shapes are not the same.",
            "  Key: data1",
            "  Obtained: (10,)",
            "  Expected: (11,)",
        ]
    )
    assert expected in obtained_error_msg


def test_arrays_of_same_shape(ndarrays_regression):
    data = {
        "2d": np.zeros((3, 4), dtype=int),
        "3d": np.ones((7, 8, 9), dtype=float),
        "4d": np.full((2, 1, 1, 4), 3, dtype=int),
    }
    ndarrays_regression.check(data)


def test_arrays_with_different_shapes(ndarrays_regression):
    # Prepare data with one shape.
    data = {"2d": np.zeros((3, 4), dtype=int)}
    ndarrays_regression.check(data)

    # Check with other shape.
    data = {"2d": np.zeros((3, 2), dtype=int)}
    with pytest.raises(AssertionError) as excinfo:
        ndarrays_regression.check(data)
    obtained_error_msg = str(excinfo.value)
    expected = "\n".join(
        [
            "Shapes are not the same.",
            "  Key: 2d",
            "  Obtained: (3, 2)",
            "  Expected: (3, 4)",
        ]
    )
    assert expected in obtained_error_msg


def test_scalars(ndarrays_regression):
    # Initial data with scalars.
    data1 = {"data1": 4.0, "data2": 42}
    ndarrays_regression.check(data1)

    # Run check with non-scalar data.
    data2 = {"data1": np.array([4.0]), "data2": np.array([42, 21])}
    with pytest.raises(AssertionError) as excinfo:
        ndarrays_regression.check(data2)
    obtained_error_msg = str(excinfo.value)
    expected = "\n".join(
        [
            "Shapes are not the same.",
            "  Key: data1",
            "  Obtained: (1,)",
            "  Expected: ()",
        ]
    )
    assert expected in obtained_error_msg

    # Other test case.
    data3 = {"data1": 5.0, "data2": 21}
    with pytest.raises(AssertionError) as excinfo:
        ndarrays_regression.check(data3)
    obtained_error_msg = str(excinfo.value)
    expected = "\n".join(
        [
            "data1:",
            "  Shape: ()",
            "  Number of differences: 1 / 1 (100.0%)",
            "  Individual errors:",
            "          Index              Obtained              Expected            Difference",
            "             ()                   5.0                   4.0                   1.0",
        ]
    )
    assert expected in obtained_error_msg
    expected = "\n".join(
        [
            "data2:",
            "  Shape: ()",
            "  Number of differences: 1 / 1 (100.0%)",
            "  Individual errors:",
            "          Index              Obtained              Expected            Difference",
            "             ()                    21                    42                   -21",
        ]
    )
    assert expected in obtained_error_msg


def test_string_array(ndarrays_regression):
    # Initial data.
    data1 = {"potato": ["delicious", "nutritive", "yummy"]}
    ndarrays_regression.check(data1)

    # Run check with wrong data.
    data1 = {"potato": ["delicious", "nutritive", "yikes"]}
    with pytest.raises(AssertionError) as excinfo:
        ndarrays_regression.check(data1)
    obtained_error_msg = str(excinfo.value)
    expected = "\n".join(
        [
            "potato:",
            "  Shape: (3,)",
            "  Number of differences: 1 / 3 (33.3%)",
            "  Individual errors:",
            "          Index              Obtained              Expected            Difference",
            "              2                 yikes                 yummy                      ",
        ]
    )
    assert expected in obtained_error_msg

    # Try data with incompatible dtype
    data1 = {"potato": ["disgusting", "nutritive", "yikes"]}
    with pytest.raises(AssertionError) as excinfo:
        ndarrays_regression.check(data1)
    obtained_error_msg = str(excinfo.value)
    expected = "\n".join(
        [
            "potato:",
            "  Shape: (3,)",
            "  Number of differences: 2 / 3 (66.7%)",
            "  Individual errors:",
            "          Index              Obtained              Expected            Difference",
            "              0            disgusting             delicious                      ",
            "              2                 yikes                 yummy                      ",
        ]
    )
    assert expected in obtained_error_msg


def test_non_dict(ndarrays_regression):
    data = np.ones(shape=(10, 10))
    with pytest.raises(TypeError) as excinfo:
        ndarrays_regression.check(data)
    obtained_error_msg = str(excinfo.value)
    expected = "\n".join(
        [
            "Only dictionaries with NumPy arrays or array-like objects are supported "
            "on ndarray_regression fixture.",
            "Object with type '{}' was given.".format(
                str(type(data)),
            ),
        ]
    )
    assert expected in obtained_error_msg


def test_structured_array(ndarrays_regression):
    data = {
        "array": np.array(
            [("spam", 1, 3.0), ("egg", 0, 4.3)],
            dtype=[("item", "U5"), ("count", "i4"), ("price", "f8")],
        )
    }
    with pytest.raises(TypeError) as excinfo:
        ndarrays_regression.check(data)
    obtained_error_msg = str(excinfo.value)
    expected = "\n".join(
        [
            "Only numeric or unicode data is supported on ndarrays_regression fixture.",
            "Array 'array' with type '{}' was given.".format(data["array"].dtype),
        ]
    )
    assert expected in obtained_error_msg


def test_new_obtained(ndarrays_regression):
    # Prepare data with one array.
    data = {"ar1": np.array([2.3, 9.4])}
    ndarrays_regression.check(data)

    # Run check with two arrays.
    data = {"ar1": np.array([2.3, 9.4]), "ar2": np.array([3, 4, 9])}
    with pytest.raises(AssertionError) as excinfo:
        ndarrays_regression.check(data)
    obtained_error_msg = str(excinfo.value)
    expected = "\n".join(
        [
            "They keys in the obtained results differ from the expected results.",
            "  Matching keys: ['ar1']",
            "  New in obtained: ['ar2']",
            "  Missing from obtained: []",
        ]
    )
    assert expected in obtained_error_msg


def test_missing_obtained(ndarrays_regression):
    # Prepare data with two arrays.
    data = {"ar1": np.array([2.3, 9.4]), "ar2": np.array([3, 4, 9])}
    ndarrays_regression.check(data)

    # Run check with just one array.
    data = {"ar1": np.array([2.3, 9.4])}
    with pytest.raises(AssertionError) as excinfo:
        ndarrays_regression.check(data)
    obtained_error_msg = str(excinfo.value)
    expected = "\n".join(
        [
            "They keys in the obtained results differ from the expected results.",
            "  Matching keys: ['ar1']",
            "  New in obtained: []",
            "  Missing from obtained: ['ar2']",
        ]
    )
    assert expected in obtained_error_msg


@pytest.mark.parametrize("prefix", [True, False])
def test_corrupt_npz(ndarrays_regression, tmp_path, prefix):
    data = {"data1": np.array([4, 5])}
    fn_npz = tmp_path / "corrupt.npz"
    # Write random bytes to a file
    with open(fn_npz, "wb") as f:
        if prefix:
            f.write(b"PK\x03\x04")
        np.random.randint(0, 256, 1000, dtype=np.ubyte).tofile(f)
    with pytest.raises(IOError) as excinfo:
        ndarrays_regression.check(data, fullpath=fn_npz)
    obtained_error_msg = str(excinfo.value)
    expected = f"NPZ file {fn_npz} could not be loaded. Corrupt file?"
    assert expected in obtained_error_msg



import os
from pathlib import Path
from typing import Any
from typing import Dict
from typing import Optional

import pytest

from .common import import_error_message
from .common import perform_regression_check


class DataFrameRegressionFixture:
    """
    Pandas DataFrame Regression fixture implementation used on dataframe_regression fixture.
    """

    DISPLAY_PRECISION = 17  # Decimal places
    DISPLAY_WIDTH = 1000  # Max. Chars on outputs
    DISPLAY_MAX_COLUMNS = 1000  # Max. Number of columns (see #3)

    def __init__(
        self, datadir: Path, original_datadir: Path, request: pytest.FixtureRequest
    ) -> None:
        self._tolerances_dict: Dict[str, Dict[str, float]] = {}
        self._default_tolerance: Dict[str, float] = {}

        self.request = request
        self.datadir = datadir
        self.original_datadir = original_datadir
        self._force_regen = False
        self._with_test_class_names = False

        self._pandas_display_options = (
            "display.precision",
            DataFrameRegressionFixture.DISPLAY_PRECISION,
            "display.width",
            DataFrameRegressionFixture.DISPLAY_WIDTH,
            "display.max_columns",
            DataFrameRegressionFixture.DISPLAY_MAX_COLUMNS,
        )

    def _check_data_types(
        self, key: Any, obtained_column: Any, expected_column: Any
    ) -> None:
        """
        Check if data type of obtained and expected columns are the same. Fail if not.
        Helper method used in _check_fn method.
        """
        try:
            import numpy as np
        except ModuleNotFoundError:
            raise ModuleNotFoundError(import_error_message("NumPy"))

        __tracebackhide__ = True
        obtained_data_type = obtained_column.values.dtype
        expected_data_type = expected_column.values.dtype
        if obtained_data_type != expected_data_type:
            # Check if both data types are comparable as numbers (float, int, short, bytes, etc...)
            if np.issubdtype(obtained_data_type, np.number) and np.issubdtype(
                expected_data_type, np.number
            ):
                return

            # In case they are not, assume they are not comparable
            error_msg = (
                "Data type for data %s of obtained and expected are not the same.\n"
                "Obtained: %s\n"
                "Expected: %s\n" % (key, obtained_data_type, expected_data_type)
            )
            raise AssertionError(error_msg)

    def _check_data_shapes(self, obtained_column: Any, expected_column: Any) -> None:
        """
        Check if obtained and expected columns have the same size.
        Helper method used in _check_fn method.
        """
        __tracebackhide__ = True

        obtained_data_shape = obtained_column.values.shape
        expected_data_shape = expected_column.values.shape
        if obtained_data_shape != expected_data_shape:
            error_msg = (
                "Obtained and expected data shape are not the same.\n"
                "Obtained: %s\n"
                "Expected: %s\n" % (obtained_data_shape, expected_data_shape)
            )
            raise AssertionError(error_msg)

    def _check_fn(self, obtained_filename: Path, expected_filename: Path) -> None:
        """
        Check if dict contents dumped to a file match the contents in expected file.
        """
        try:
            import numpy as np
        except ModuleNotFoundError:
            raise ModuleNotFoundError(import_error_message("NumPy"))
        try:
            import pandas as pd
        except ModuleNotFoundError:
            raise ModuleNotFoundError(import_error_message("Pandas"))

        __tracebackhide__ = True

        obtained_data = pd.read_csv(str(obtained_filename))
        expected_data = pd.read_csv(str(expected_filename))

        comparison_tables_dict = {}
        for k in obtained_data.keys():
            obtained_column = obtained_data[k]
            expected_column = expected_data.get(k)

            if expected_column is None:
                error_msg = f"Could not find key '{k}' in the expected results.\n"
                error_msg += "Keys in the obtained data table: ["
                for k in obtained_data.keys():
                    error_msg += f"'{k}', "
                error_msg += "]\n"
                error_msg += "Keys in the expected data table: ["
                for k in expected_data.keys():
                    error_msg += f"'{k}', "
                error_msg += "]\n"
                error_msg += "To update values, use --force-regen option.\n\n"
                raise AssertionError(error_msg)

            tolerance_args = self._tolerances_dict.get(k, self._default_tolerance)

            self._check_data_types(k, obtained_column, expected_column)
            self._check_data_shapes(obtained_column, expected_column)

            if np.issubdtype(obtained_column.values.dtype, np.inexact):
                not_close_mask = ~np.isclose(
                    obtained_column.values,
                    expected_column.values,
                    equal_nan=True,
                    **tolerance_args,
                )
            else:
                not_close_mask = obtained_column.values != expected_column.values
                # If Empty/NaN data is expected, then the values are equal:
                not_close_mask[
                    np.logical_and(
                        pd.isna(obtained_column.values), pd.isna(expected_column.values)
                    )
                ] = False

            if np.any(not_close_mask):
                diff_ids = np.where(not_close_mask)[0]
                diff_obtained_data = obtained_column[diff_ids]
                diff_expected_data = expected_column[diff_ids]
                if obtained_column.values.dtype == bool:
                    diffs = np.logical_xor(obtained_column, expected_column)[diff_ids]
                elif obtained_column.values.dtype == object:
                    diffs = diff_obtained_data.copy()
                    diffs[:] = "?"
                else:
                    diffs = np.abs(obtained_column - expected_column)[diff_ids]

                comparison_table = pd.concat(
                    [diff_obtained_data, diff_expected_data, diffs], axis=1
                )
                comparison_table.columns = [f"obtained_{k}", f"expected_{k}", "diff"]
                comparison_tables_dict[k] = comparison_table

        if len(comparison_tables_dict) > 0:
            error_msg = "Values are not sufficiently close.\n"
            error_msg += "To update values, use --force-regen option.\n\n"
            for k, comparison_table in comparison_tables_dict.items():
                error_msg += f"{k}:\n{comparison_table}\n\n"
            if obtained_column.values.dtype == object:
                error_msg += (
                    "WARNING: diffs for this kind of data type cannot be computed."
                )
            raise AssertionError(error_msg)

    def _dump_fn(self, data_object: Any, filename: Path) -> None:
        """
        Dump dict contents to the given filename
        """
        data_object.to_csv(
            str(filename),
            float_format=f"%.{DataFrameRegressionFixture.DISPLAY_PRECISION}g",
        )

    def check(
        self,
        data_frame: Any,
        basename: Optional[str] = None,
        fullpath: Optional["os.PathLike[str]"] = None,
        tolerances: Optional[Dict[str, Dict[str, float]]] = None,
        default_tolerance: Optional[Dict[str, float]] = None,
    ) -> None:
        """
        Checks a pandas dataframe, containing only numeric data, against a previously recorded version, or generate a new file.

        Example::

            data_frame = pandas.DataFrame.from_dict({
                'U_gas': U[0][positions],
                'U_liquid': U[1][positions],
                'gas_vol_frac [-]': vol_frac[0][positions],
                'liquid_vol_frac [-]': vol_frac[1][positions],
                'P': Pa_to_bar(P)[positions],
            })
            dataframe_regression.check(data_frame)

        :param data_frame: pandas DataFrame containing data for regression check.

        :param basename: basename of the file to test/record. If not given the name
            of the test is used.

        :param fullpath: complete path to use as a reference file. This option
            will ignore embed_data completely, being useful if a reference file is located
            in the session data dir for example.

        :param tolerances: dict mapping keys from the data_frame to tolerance settings for the
            given data. Example::

                tolerances={'U': dict(atol=1e-2)}

        :param default_tolerance: dict mapping the default tolerance for the current check
            call. Example::

                default_tolerance=dict(atol=1e-7, rtol=1e-18).

            If not provided, will use defaults from numpy's ``isclose`` function.

        ``basename`` and ``fullpath`` are exclusive.
        """
        try:
            import pandas as pd
        except ModuleNotFoundError:
            raise ModuleNotFoundError(import_error_message("Pandas"))

        import functools

        __tracebackhide__ = True

        assert isinstance(data_frame, pd.DataFrame), (
            "Only pandas DataFrames are supported on dataframe_regression fixture.\n"
            "Object with type '%s' was given." % (str(type(data_frame)),)
        )

        for column in data_frame.columns:
            array = data_frame[column]
            # Skip assertion if an array of strings
            if (array.dtype == "O") and (type(array.iloc[0]) is str):
                continue
            # Rejected: timedelta, datetime, objects, zero-terminated bytes, unicode strings and raw data
            assert array.dtype not in ["m", "M", "O", "S", "a", "U", "V"], (
                "Only numeric data is supported on dataframe_regression fixture.\n"
                "Array with type '%s' was given." % (str(array.dtype),)
            )

        if tolerances is None:
            tolerances = {}
        self._tolerances_dict = tolerances

        if default_tolerance is None:
            default_tolerance = {}
        self._default_tolerance = default_tolerance

        dump_fn = functools.partial(self._dump_fn, data_frame)

        with pd.option_context(*self._pandas_display_options):
            perform_regression_check(
                datadir=self.datadir,
                original_datadir=self.original_datadir,
                request=self.request,
                check_fn=self._check_fn,
                dump_fn=dump_fn,
                extension=".csv",
                basename=basename,
                fullpath=fullpath,
                force_regen=self._force_regen,
                with_test_class_names=self._with_test_class_names,
            )



import os
from functools import partial
from pathlib import Path
from typing import Callable
from typing import Optional
from typing import Union

import pytest

from .common import check_text_files
from .common import perform_regression_check


class FileRegressionFixture:
    """
    Implementation of `file_regression` fixture.
    """

    def __init__(
        self, datadir: Path, original_datadir: Path, request: pytest.FixtureRequest
    ) -> None:
        self.request = request
        self.datadir = datadir
        self.original_datadir = original_datadir
        self.force_regen = False
        self.with_test_class_names = False

    def check(
        self,
        contents: Union[str, bytes],
        encoding: Optional[str] = None,
        extension: str = ".txt",
        newline: Optional[str] = None,
        basename: Optional[str] = None,
        fullpath: Optional["os.PathLike[str]"] = None,
        binary: bool = False,
        obtained_filename: Optional["os.PathLike[str]"] = None,
        check_fn: Optional[Callable[[Path, Path], None]] = None,
    ) -> None:
        """
        Checks the contents against a previously recorded version, or generate a new file.

        :param contents: content to be verified.
        :param encoding: Encoding used to write file, if any.
        :param extension: Extension of file.
        :param newline: See `io.open` docs.
        :param binary: If the file is binary or text.
        :param obtained_filename: ..see:: FileRegressionCheck
        :param check_fn: a function with signature ``(obtained_filename, expected_filename)`` that should raise
            AssertionError if both files differ.
            If not given, use internal function which compares text using :py:mod:`difflib`.
        """
        __tracebackhide__ = True

        if binary and encoding:
            raise ValueError(
                "Only binary ({!r}) or encoding ({!r}) parameters must be passed at the same time.".format(
                    binary, encoding
                )
            )

        if binary:
            assert isinstance(
                contents, bytes
            ), "Expected bytes contents but received type {}".format(
                type(contents).__name__
            )
        else:
            assert isinstance(
                contents, str
            ), "Expected text/unicode contents but received type {}".format(
                type(contents).__name__
            )

        if check_fn is None:
            if binary:

                def check_fn(obtained_filename: Path, expected_filename: Path) -> None:
                    if obtained_filename.read_bytes() != expected_filename.read_bytes():
                        raise AssertionError(
                            "Binary files {} and {} differ.".format(
                                obtained_filename, expected_filename
                            )
                        )

            else:
                check_fn = partial(check_text_files, encoding=encoding)

        def dump_fn(filename: Path) -> None:
            mode = "wb" if binary else "w"
            with open(str(filename), mode, encoding=encoding, newline=newline) as f:
                f.write(contents)

        assert check_fn is not None
        perform_regression_check(
            datadir=self.datadir,
            original_datadir=self.original_datadir,
            request=self.request,
            check_fn=check_fn,
            dump_fn=dump_fn,
            extension=extension,
            basename=basename,
            fullpath=fullpath,
            force_regen=self.force_regen,
            with_test_class_names=self.with_test_class_names,
            obtained_filename=obtained_filename,
        )

    # non-PEP 8 alias used internally at ESSS
    Check = check
